/* busarbc */
/* ############################################################### */
/* ############################################################### */
/* ###                                                         ### */
/* ###  IOMD DMA ARBITER block level model                     ### */
/* ###                                                         ### */
/* ###  Created 03/9/92: David Flynn                           ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ############################################################### */
/* ############################################################### */
/* ###  Revision History:                                      ### */
/* ###   03/9/92: DWF : partion into separate blocks           ### */
/* ###   07/9/92: DWF : dmarq sticky until dmaend              ### */
/* ###   08/9/92: DWF : dmarq arbitration RCLK falling sync    ### */
/* ###   09/9/92: DWF : <Nreset> added to BUSARB to clear DMAs ### */
/* ###   29/9/92L AS  : Added dbe output                       ### */
/* ###   29/9/92L DWF : added sound 0 channel arbitration      ### */
/* ###  10/11/92: DWF : remove IOGOING inhibit on IODRQ arbtn. ### */
/* ###  11/12/92: DWF : remove IOGOING altogether!             ### */
/*      25/1/93   WHO   added new output: Nuldmarq                 */
/* ############################################################### */


#include <Handy.h>
#include "busarb.h"

/* ############################################################### */

typedef struct
{
  ModelCore     model;
  Signal        *dummy;

} *DelayModel;

int get_delay (delay_name, model)
char *delay_name;
VoidStar model;
{
  Parameter    param;
  int          delay = 1000*NS;
  DelayModel   me = (DelayModel)model;

  param = simmodel_param(me->model.state,"delay");
  if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  if (simmodel_type(param) == PT_PARAMBLOCK) {
    param = simmodel_sub_param(param,delay_name);
    if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  }
  errhandler_f (SS_MODELS+DISASTER,"!!! %s delay not correctly specified\n",delay_name);
  return delay;
}


/* ############################################################### */
/* ############################################################### */
/* ### BUSARB : IOMD DMA channel arbiter                       ### */
/* ############################################################### */

#define BUSARB_def \
"BUSARB( \
  rclk, \
  Nreset, \
  dmadone, \
  dmagoing, \
  curdrq, \
  vsnddrq, \
  snd0drq, \
  trandrq, \
  snd1drq, \
  refdrq, \
  iodrq, \
  snd0cod, \
  dmarq, \
  Nuldmarq, \
  dmach[2:0], \
  dmavbus, \
  dmabus )"

typedef struct
{
    ModelCore   model;
    Signal  *rclk;
    Signal  *Nreset;
    Signal  *dmadone;
    Signal  *dmagoing;
    Signal  *curdrq;
    Signal  *vsnddrq;
    Signal  *snd0drq;
    Signal  *trandrq;
    Signal  *snd1drq;
    Signal  *refdrq;
    Signal  *iodrq;
    Signal  *snd0cod;
    Signal  *dmarq;
    Signal  *Nuldmarq;
    Bundle  *dmach;
    Signal  *dmavbus;
    Signal  *dmabus;
} *BUSARB_model;

typedef struct
{
     int  dt;
} BUSARB_state;

static void BUSARB_proc (modelInst)
VoidStar modelInst;
{
   BUSARB_model me = (BUSARB_model) modelInst;
   BUSARB_state *s = (BUSARB_state *)me->model.state;
   int msk;
/* firstly provide diagnostic warnings for requests that change
   during RCLK high, when the arbiter should be stabilising       */
   if (HI(rclk)) {
     if (CHANGING(curdrq))
       errhandler_f(0,"WARNING: curdrq CHANGING in RCLK hi-phase\n");
     if (CHANGING(vsnddrq))
       errhandler_f(0,"WARNING: vsnddrq CHANGING in RCLK hi-phase\n");
     if (CHANGING(snd0drq))
       errhandler_f(0,"WARNING: snd0drq CHANGING in RCLK hi-phase\n");
     if (CHANGING(trandrq))
       errhandler_f(0,"WARNING: trandrq CHANGING in RCLK hi-phase\n");
     if (CHANGING(snd1drq))
       errhandler_f(0,"WARNING: snd1drq CHANGING in RCLK hi-phase\n");
     if (CHANGING(refdrq))
       errhandler_f(0,"WARNING: refdrq CHANGING in RCLK hi-phase\n");
     if (CHANGING(iodrq))
       errhandler_f(0,"WARNING: iodrq CHANGING in RCLK hi-phase\n");
   }
/* priority encode (highest takes precedence) */
   msk = NODMA;  /* default case */
/* written as latest in list takes precedence */
   if (HI(iodrq))  msk = IODMA;
   if (HI(refdrq))  msk = REFDMA;
   if (HI(snd1drq)) msk = SND1DMA;
   if (HI(trandrq)) msk = TRANDMA;
 /* source of snd0dma depends on state of snd0cod input */
   if (HI(snd0cod) && HI(snd0drq)) msk = SND0DMA;
   if (LO(snd0cod) && HI(vsnddrq)) msk = SND0DMA;
   if (HI(curdrq))  msk = CURSDMA;
/* overiding reset clears all dma requests */
   if (LO(Nreset)) {
       SignalEvt(dmarq, ZERO, s->dt);
       BundleEvt(dmach, NODMA, s->dt);
       SignalEvt(dmavbus, ZERO, s->dt);
   } else {
  /* dma requests synchronised on falling RCLK and metastable settling... */
     if (HI(rclk)) {
  /* transparent 'clean' dmarq issued during RCLK high */
       if ((msk != NODMA) || (HI(dmagoing) && LO(dmadone)))
         SignalEvt(dmarq, ONE, s->dt);
       else
         SignalEvt(dmarq, ZERO, s->dt);
     }
  /* output DMA channel number: effectively latched for duration of dma
     only change on falling RCLK
     IFF not doing dma OR further dma and dmadone signal)
  */
     if (FALLING(rclk) && (LO(dmagoing) || HI(dmadone))) {
       BundleEvt(dmach, msk, s->dt);
       if (((msk==SND0DMA) && LO(snd0cod)) || (msk==CURSDMA))
         SignalEvt(dmavbus, ONE, s->dt);
       else
         SignalEvt(dmavbus, ZERO, s->dt);
       if ((msk==NODMA) || (msk==REFDMA) || (msk==TRANDMA))
         SignalEvt(dmabus, ZERO, s->dt);
       else SignalEvt(dmabus, ONE, s->dt);
     }
   } /* end else */

   if ((msk != NODMA) || (HI(dmagoing) && LO(dmadone)))
      SignalEvt(Nuldmarq, ZERO, s->dt);
   else
      SignalEvt(Nuldmarq,  ONE, s->dt);
}


static void BUSARB_init (modelInst)
VoidStar modelInst;
{
    BUSARB_state *s;
    BUSARB_model me = (BUSARB_model) modelInst;
    NEW(s, BUSARB_state);
    me->model.proc = BUSARB_proc;
    if (!s) errhandler_f(SS_MODELS+DISASTER,"Malloc failure\n");
    s->dt = get_delay("dt",me);
    me->model.state = (VoidStar ) s;
}

void BUSARBC_declare()
{
    simmodel_begin("iomd.busarbc");
    simmodel_declare_s(BUSARB_def, BUSARB_init);
    simmodel_end();
}
